@@ -1,5 +1,5 @@
 # -*- encoding: utf-8
-# Copyright (C) 2009  Rene Koecher <shirk87@gmail.com>
+# Copyright (C) 2009-2011  Rene Koecher <shirk@bitspin.org>
 #
 # This file is part of paraproxy.
 #
@@ -28,6 +28,8 @@ import thread
 import fnmatch
 import getpass
 import subprocess
+import shlex
+import StringIO
 
 from WrapSock import WrapSock
 
@@ -44,8 +46,8 @@ try:
 except AttributeError:
 	__paramiko_SSHClient_delete_hook = None
 
-# list with configured proxy-commands
-__paramiko_SSHClient_proxy_list     = {}
+# list with configured proxy-commands / ssh option etc.
+__paramiko_SSHClient_proxy_config = paramiko.SSHConfig()
 
 # Import paramiko internals with prefix (__.*)
 from paramiko.hostkeys import HostKeys as __HostKeys
@@ -57,6 +59,62 @@ def __paramiko_SSHClient_debug(msg):
 	if __paramiko_SSHClient_connect_verbose:
 		print '[%s] %s' % (os.getpid(), msg)
 
+def __paramiko_SSHClient_proxy_cmdchain(self, hostname, rec=0):
+	"""
+	Resolver used to detect multi-layer ProxyCommand chains
+
+	This will search for hostname in proxy_list, if found
+	add the assigned ProxyCommand to the chain and check if it it turn
+	requires a ProxyCommand
+
+	Chain detection depends on the command to be 'ssh'/'ssh.exe' (absolute paths are supported)
+	"""
+	chain = []
+	command = __paramiko_SSHClient_proxy_config.lookup(hostname).get('proxycommand')
+	port = __paramiko_SSHClient_proxy_config.lookup(hostname).get('port', '22')
+	hostname = __paramiko_SSHClient_proxy_config.lookup(hostname).get('hostname', hostname)
+	if not command:
+		if rec != 0:
+			__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): hop_%d: %s:%s, DIRECT' % (rec,hostname,port))
+			return [(hostname, 'DIRECT')]
+		return []
+
+	# command found
+	try:
+		#__paramiko_SSHClient_debug('processing "%s"' % command)
+		command_args = [] + shlex.split(command)
+
+		exe = os.path.basename(command_args[0].lower())
+		if exe.lower() == 'ssh' or exe.lower() == 'ssh.exe':
+			# command is SSH, extract the host and the remote command
+			remote_cmd = command_args.pop()
+			remote_host = command_args.pop()
+
+			chain = [(remote_host, remote_cmd.replace('%h', hostname).replace('%p', port))]
+			next_hop = __paramiko_SSHClient_proxy_cmdchain(self, remote_host, rec+1)
+
+			if next_hop:
+				chain = next_hop + chain
+		else:
+			# a regular command like netcat
+			chain = [(hostname, command.replace('%h', hostname).replace('%p', port))]
+			if rec != 0:
+			    __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): WARNING: regular command "%s" for "%s" will be run in hop_%d'  % (command, hostname, rec))
+			    __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): WARNING: paraproxy doesn\'t know how to chain further from here.')
+			return chain
+	except Exception,e:
+		#FIXME: propper handling!
+		 __paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): ERROR: failed to parse ProxyCommand for host "%s" - abort!' % (hostname))
+		 raise
+
+	if rec != 0:
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): hop_%d: %s:%s' % (rec,hostname, port))
+	else:
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): target: %s:%s' % (hostname, port))
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_proxy_cmdchain(): final chain: %s' % (' => '.join([('(%s, %s)' % (x[0],x[1])) for x in chain])))
+
+	return chain
+
 def __paramiko_SSHClient_proxy_thread(self, ip, port, command):
 	"""
 	Proxy thread used to establish a connection using a ssh ProxyCommand.
@@ -82,11 +140,13 @@ def __paramiko_SSHClient_proxy_thread(self, ip, port, command):
 	(stmp, addr) = sock.accept()
 	__paramiko_SSHClient_debug('proxy: incomming connection')
 	
-	wrap_sock = WrapSock(stmp)
 	cmd_str = str(command) # copy
 	self._proc_id = None
+	wrap_sock = WrapSock(stmp)
 
 	try:
+		__paramiko_SSHClient_debug('proxy: executing command "%s"' % cmd_str.replace('%h', str(ip)).replace('%p', str(port)))
+
 		proc = subprocess.Popen(cmd_str.replace('%h', str(ip)).replace('%p', str(port)),
 								shell=True,
 								stdin=wrap_sock,
@@ -94,7 +154,8 @@ def __paramiko_SSHClient_proxy_thread(self, ip, port, command):
 								close_fds=True)
 	
 		self._proc_id = proc.pid
-	
+		self._sock_state = 1
+
 		# warten auf Beendigung der Verbindung
 		while proc.poll() == None \
 		      and not self._should_terminate:
@@ -132,62 +193,39 @@ def __paramiko_SSHClient_connect(self, hostname, port=22, username=None,
 	Hook function to enable Paramiko's SSHClient to work with ssh ProxyCommand.
 	This will only hook into connections matching one of the ProxyRules.
 	"""
-	__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(%s, %s)' % (hostname, port))
-	
-	command = None
-	for pattern in __paramiko_SSHClient_proxy_list.keys():
-		if fnmatch.fnmatch(hostname, pattern):
-			command = __paramiko_SSHClient_proxy_list[pattern]
-			break
-	
-	if not command:
-		__paramiko_SSHClient_debug('direct connection.')
-		return __paramiko_SSHClient_connect_hook(self, hostname, port, username, password, pkey,
-							 key_filename, timeout, allow_agent, look_for_keys)
 
-	__paramiko_SSHClient_debug('using ProxyCommand `%s`' % command)
-	
-	self._sock_name = None
-	self._should_terminate = False
+	def __create_and_connect_transport(self, _hostname, _transport_sock, _username=None, _password=None):
+		"""
+		Helper function to create a new Transport and perform the required SSH authentication
+		-- just here to keep the code more readable
+		"""
+		#print 'Create transport on: %s' % _transport_sock
+		t = self._transport = __Transport(_transport_sock)
+		self._chained_transports.append(t)
 
-	thread.start_new(__paramiko_SSHClient_proxy_thread, (self, hostname, port, command))
-	time.sleep(1)
-	
-	sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-	if timeout is not None:
-		try:
-			sock.settimeout(timeout)
-		except:
-			pass
-
-	try:
-		sock.connect(self._sock_name) # set by proxy_thread
-		__paramiko_SSHClient_debug('connected to proxy thread on `%s`' % self._sock_name)
-	
-		t = self._transport = __Transport(sock)
-	
 		if self._log_channel is not None:
 			t.set_log_channel(self._log_channel)
+
 		t.start_client()
 		__ResourceManager.register(self, t)
-	
+		
 		server_key = t.get_remote_server_key()
 		keytype = server_key.get_name()
-	
-		our_server_key = self._system_host_keys.get(hostname, {}).get(keytype, None)
+		
+		our_server_key = self._system_host_keys.get(_hostname, {}).get(keytype, None)
 		if our_server_key is None:
-			our_server_key = self._host_keys.get(hostname, {}).get(keytype, None)
+			our_server_key = self._host_keys.get(_hostname, {}).get(keytype, None)
 		if our_server_key is None:
 			# will raise exception if the key is rejected; let that fall out
-			self._policy.missing_host_key(self, hostname, server_key)
+			self._policy.missing_host_key(self, _hostname, server_key)
 			# if the callback returns, assume the key is ok
 			our_server_key = server_key
 	
 		if server_key != our_server_key:
-			raise BadHostKeyException(hostname, server_key, our_server_key)
+			raise BadHostKeyException(_hostname, server_key, our_server_key)
 	
-		if username is None:
-			username = getpass.getuser()
+		if _username is None:
+			_username = getpass.getuser()
 	
 		if key_filename is None:
 			key_filenames = []
@@ -195,8 +233,89 @@ def __paramiko_SSHClient_connect(self, hostname, port=22, username=None,
 			key_filenames = [ key_filename ]
 		else:
 			key_filenames = key_filename
-		self._auth(username, password, pkey, key_filenames, allow_agent, look_for_keys)
+		self._auth(_username, _password, pkey, key_filenames, allow_agent, look_for_keys)
+	#---
 
+	__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(%s, %s)' % (hostname, port))
+	command = __paramiko_SSHClient_proxy_config.lookup(hostname).get('proxycommand')
+	
+	if not command:
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): direct connection.')
+		return __paramiko_SSHClient_connect_hook(self, hostname, port, username, password, pkey,
+							 key_filename, timeout, allow_agent, look_for_keys)
+
+	command_chain = __paramiko_SSHClient_proxy_cmdchain(self, hostname)
+	if len(command_chain) > 1:
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): using multi-layer chain with %d hops' % (len(command_chain) - 1))
+	else:
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): using ProxyCommand `%s`' % command)
+
+	self._sock_name = ''
+	self._sock_state = 0
+	self._proc_id = 0
+	self._should_terminate = False
+	self._chained_transports = []
+
+	if not command_chain:
+		thread.start_new(__paramiko_SSHClient_proxy_thread, (self, hostname, port, command))
+		time.sleep(1)
+		
+		sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+		if timeout is not None:
+			try:
+				sock.settimeout(timeout)
+			except:
+				pass
+
+		try:
+			sock.connect(self._sock_name) # set by proxy_thread
+			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to proxy thread on `%s`' % self._sock_name)
+			while self._sock_state == 0:
+				time.sleep(0.5)
+
+		except Exception, e:
+			# except *anything* to be able to remove the proxy thread.
+			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): caught %s - closing threads.' % e.__class__.__name__)
+
+			try:
+				if (self._proc_id):
+					__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): requesting thread shutdown')
+					self._should_terminate = True
+				else:
+					__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): os._proc_id <= 0, nothing to do.')
+			except OSError:
+				pass
+
+			raise
+
+	else:
+		(proxy_host, proxy_command) = command_chain.pop(0)
+		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+		if timeout is not None:
+			try:
+				sock.settimeout(timeout)
+			except:
+				pass
+
+		# no try: except: here - just raise any socke errors
+		proxy_port = long(__paramiko_SSHClient_proxy_config.lookup(proxy_host).get('port', '22'))
+		sock.connect((proxy_host, proxy_port))
+		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to `%s:%s`' % (proxy_host, proxy_port))
+
+	try:
+		transport_sock = sock
+		while command_chain:
+			__create_and_connect_transport(self, command_chain[0][0], transport_sock, username, password)
+			# this hop is now valid, execute the ProxyCommand
+			(proxy_host, proxy_command) = command_chain.pop(0)
+
+			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): connected to %s, sending command: %s' % (proxy_host, proxy_command))
+			transport_sock = self._transport.open_session()
+			__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): next transport_sock: %s' % transport_sock)
+			transport_sock.exec_command(proxy_command)
+
+		# final connection, always executed
+		__create_and_connect_transport(self, hostname, transport_sock, username, password)
 	except Exception, e:
 		# except *anything* to be able to remove the proxy thread.
 		__paramiko_SSHClient_debug('__paramiko_SSHClient_connect(): caught %s - closing threads.' % e.__class__.__name__)
@@ -215,6 +334,17 @@ def __paramiko_SSHClient_connect(self, hostname, port=22, username=None,
 	__paramiko_SSHClient_debug('~__paramiko_SSHClient_connect()')
 
 def __paramiko_SSHClient_close(self):
+	__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): begin closing.')
+	try:
+		if self._chained_transports:
+			self._transport = None
+			while self._chained_transports:
+				__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): remove hop transport..')
+				t = self._chained_transports.pop(0).close()
+				del t
+	except AttributeError:
+	    pass
+
 	__paramiko_SSHClient_close_hook(self)
 	try:
 		while self._proc_id:
@@ -233,8 +363,20 @@ def __paramiko_SSHClient_close(self):
 	except AttributeError:
 		pass
 
+	__paramiko_SSHClient_debug('__paramiko_SSHClient_close(): finished closing.')
+
 def __paramiko_SSHClient_delete(self):
 	__paramiko_SSHClient_debug('__paramiko_SSHClient_delete(): begin destruction.')
+	try:
+		if self._chained_transports:
+			self._transport = None
+			while self._chained_transports:
+				__paramiko_SSHClient_debug('__paramiko_SSHClient_delete(): remove hop transport..')
+				t = self._chained_transports.pop(0).close()
+				del t
+	except AttributeError:
+	    pass
+
 	try:
 		while self._proc_id:
 			self._should_terminate = True
@@ -258,30 +400,23 @@ def __paramiko_SSHClient_delete(self):
 	__paramiko_SSHClient_debug('~__paramiko_SSHClient_delete(): finished destruction.')
 
 # parse ssh_config
+merged_config = StringIO.StringIO()
 for conf_file in [os.path.join(os.path.sep, 'etc', 'ssh', 'ssh_config'),
 			       os.path.join(os.getenv('HOME','/tmp'), '.ssh', 'config')]:
 	try:
 		config = open(conf_file)
-		host_name = None
-		for line in config.readlines():
-			line = line.strip()
-			match = re.match(r'[ \t]*Host[ \t]+([^ \t#]+)', line)
-			if match:
-				__paramiko_SSHClient_debug('New host `%s`' % match.groups()[0])
-				host_name = match.groups()[0]
-		
-			match = re.match(r'[ \t]*ProxyCommand[ \t]+([^;#]+)', line)
-			if match:
-				__paramiko_SSHClient_debug('New command `%s`' % match.groups()[0])
-				__paramiko_SSHClient_proxy_list[host_name] = match.groups()[0]
-
+		merged_config.write(config.read())
+		merged_config.write('\n')
 		config.close()
 		del config
-		del host_name
-	
+
 	except IOError,e:
 		pass
 
+merged_config.seek(0,0)
+__paramiko_SSHClient_proxy_config.parse(merged_config)
+del merged_config
+
 # finally modify SSHClient to use our connect hook
 setattr(globals()['paramiko'].SSHClient, 'connect', __paramiko_SSHClient_connect)
 setattr(globals()['paramiko'].SSHClient, 'close'  , __paramiko_SSHClient_close)
